/**
 * @file libgalago-gtk/galago-gtk-account-option-menu.c
 *       Account OptionMenu widget
 *
 * @Copyright (C) 2004-2005 Christian Hammond.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
#include <libgalago-gtk/galago-gtk-account-option-menu.h>
#include <libgalago-gtk/galago-gtk-account-menu-item.h>
#include <libgalago-gtk/galago-gdk-pixbuf.h>
#include <libgalago/galago.h>
#include <string.h>

struct _GalagoGtkAccountOptionMenuPriv
{
	GalagoPerson *person;
	GalagoService *service;

	galago_bool me_person_set;

	gboolean me_accounts;

	GtkWidget *service_optmenu;

	gint service_changed_sig_id;
	gint service_option_menu_destroyed_sig_id;

	unsigned long account_added_id;
	unsigned long account_removed_id;
	unsigned long account_connected_id;
	unsigned long account_disconnected_id;

	unsigned long person_added_id;
	unsigned long person_destroy_id;
};

enum
{
	ACCOUNT_CHANGED,
	LAST_SIGNAL
};

enum
{
	COLUMN_DATA,
	COLUMN_ICON,
	COLUMN_NAME,
	NUM_COLUMNS
};

static void galago_gtk_account_option_menu_class_init(
	GalagoGtkAccountOptionMenuClass *klass);
static void galago_gtk_account_option_menu_init(
	GalagoGtkAccountOptionMenu *optmenu);
static void galago_gtk_account_option_menu_finalize(GObject *obj);
static void galago_gtk_account_option_menu_destroy(GtkObject *obj);
static void account_added_cb(void *obj, const char *property,
							 GalagoAccount *account,
							 GalagoGtkAccountOptionMenu *optmenu);
static void account_removed_cb(void *obj, const char *property,
							   GalagoAccount *account,
							   GalagoGtkAccountOptionMenu *optmenu);
static void person_destroy_cb(GalagoPerson *person, const char *property,
							  GalagoGtkAccountOptionMenu *optmenu);

static GtkOptionMenuClass *parent_class = NULL;
static guint signals[LAST_SIGNAL] = {0};

GType
galago_gtk_account_option_menu_get_type(void)
{
	static GType type = 0;

	if (!type)
	{
		static const GTypeInfo info =
		{
			sizeof(GalagoGtkAccountOptionMenuClass),
			NULL,
			NULL,
			(GClassInitFunc)galago_gtk_account_option_menu_class_init,
			NULL,
			NULL,
			sizeof(GalagoGtkAccountOptionMenu),
			0,
			(GInstanceInitFunc)galago_gtk_account_option_menu_init
		};

		type = g_type_register_static(GTK_TYPE_OPTION_MENU,
									  "GalagoGtkAccountOptionMenu", &info, 0);
	}

	return type;
}

static void
galago_gtk_account_option_menu_class_init(
	GalagoGtkAccountOptionMenuClass *klass)
{
	GObjectClass *gobject_class;
	GtkObjectClass *object_class;

	parent_class = g_type_class_peek_parent(klass);

	gobject_class = G_OBJECT_CLASS(klass);
	object_class  = GTK_OBJECT_CLASS(klass);

	gobject_class->finalize = galago_gtk_account_option_menu_finalize;

	object_class->destroy = galago_gtk_account_option_menu_destroy;

	signals[ACCOUNT_CHANGED] =
		g_signal_new("account_changed",
					 G_TYPE_FROM_CLASS(gobject_class),
					 G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
					 G_STRUCT_OFFSET(GalagoGtkAccountOptionMenuClass,
									 account_changed),
					 NULL, NULL,
					 gtk_marshal_VOID__POINTER,
					 G_TYPE_NONE, 1,
					 G_TYPE_POINTER);
}

static void
option_menu_changed(GalagoGtkAccountOptionMenu *optmenu, gpointer user_data)
{
	GalagoAccount *account;

	account = galago_gtk_account_option_menu_get_active(optmenu);

	g_signal_emit(optmenu, signals[ACCOUNT_CHANGED], 0, account);
}

static void
galago_gtk_account_option_menu_init(GalagoGtkAccountOptionMenu *optmenu)
{
	optmenu->priv = g_new0(GalagoGtkAccountOptionMenuPriv, 1);

	g_signal_connect(G_OBJECT(optmenu), "changed",
					 G_CALLBACK(option_menu_changed), NULL);

	optmenu->priv->account_added_id =
		galago_signal_connect_class(GALAGO_CLASS_SERVICE, "account-added",
									GALAGO_CALLBACK(account_added_cb),
									optmenu);
	optmenu->priv->account_removed_id =
		galago_signal_connect_class(GALAGO_CLASS_SERVICE, "account-removed",
									GALAGO_CALLBACK(account_removed_cb),
									optmenu);
}

static void
galago_gtk_account_option_menu_finalize(GObject *obj)
{
	GalagoGtkAccountOptionMenu *optmenu;

	g_return_if_fail(obj != NULL);
	g_return_if_fail(GALAGO_GTK_IS_ACCOUNT_OPTION_MENU(obj));

	optmenu = GALAGO_GTK_ACCOUNT_OPTION_MENU(obj);

	g_free(optmenu->priv);

	if (G_OBJECT_CLASS(parent_class)->finalize)
		G_OBJECT_CLASS(parent_class)->finalize(obj);
}

static void
galago_gtk_account_option_menu_destroy(GtkObject *obj)
{
	GalagoGtkAccountOptionMenu *optmenu;

	g_return_if_fail(obj != NULL);
	g_return_if_fail(GALAGO_GTK_IS_ACCOUNT_OPTION_MENU(obj));

	optmenu = GALAGO_GTK_ACCOUNT_OPTION_MENU(obj);

	if (optmenu->priv->service_changed_sig_id != 0)
	{
		g_signal_handler_disconnect(G_OBJECT(optmenu->priv->service_optmenu),
									optmenu->priv->service_changed_sig_id);
		optmenu->priv->service_changed_sig_id = 0;
	}

	if (optmenu->priv->service_option_menu_destroyed_sig_id != 0)
	{
		g_signal_handler_disconnect(G_OBJECT(optmenu->priv->service_optmenu),
			optmenu->priv->service_option_menu_destroyed_sig_id);
		optmenu->priv->service_option_menu_destroyed_sig_id = 0;
	}

	if (optmenu->priv->account_added_id > 0)
		galago_signals_disconnect_by_id(optmenu->priv->account_added_id);

	if (optmenu->priv->account_removed_id > 0)
		galago_signals_disconnect_by_id(optmenu->priv->account_removed_id);

	if (optmenu->priv->account_connected_id > 0)
		galago_signals_disconnect_by_id(optmenu->priv->account_connected_id);

	if (optmenu->priv->account_disconnected_id > 0)
		galago_signals_disconnect_by_id(optmenu->priv->account_disconnected_id);

	if (optmenu->priv->person_added_id > 0)
	{
		galago_signals_disconnect_by_id(optmenu->priv->person_added_id);
		optmenu->priv->person_added_id = 0;
	}

	if (optmenu->priv->person_destroy_id > 0)
	{
		galago_signals_disconnect_by_id(optmenu->priv->person_destroy_id);
		optmenu->priv->person_destroy_id = 0;
	}

	optmenu->priv->account_added_id        = 0;
	optmenu->priv->account_removed_id      = 0;
	optmenu->priv->account_connected_id    = 0;
	optmenu->priv->account_disconnected_id = 0;

	if (GTK_OBJECT_CLASS(parent_class)->destroy)
		GTK_OBJECT_CLASS(parent_class)->destroy(obj);
}

static void
append_account(GtkWidget *menu, GalagoAccount *account)
{
	GtkWidget *menuitem;

	menuitem = galago_gtk_account_menu_item_new(account);
	gtk_widget_show(menuitem);
	gtk_menu_shell_append(GTK_MENU_SHELL(menu), menuitem);
}

static void
append_accounts_from_person(GalagoGtkAccountOptionMenu *optmenu,
							GtkWidget *menu, GalagoPerson *person,
							GalagoService *service)
{
	const GalagoList *l;

	for (l = galago_person_get_accounts(person, TRUE);
		 l != NULL;
		 l = l->next)
	{
		GalagoAccount *account = (GalagoAccount *)l->data;

		if (service == NULL || galago_account_get_service(account) == service)
			append_account(menu, account);
	}
}

static void
append_accounts_from_service(GalagoGtkAccountOptionMenu *optmenu,
							 GtkWidget *menu, GalagoService *service)
{
	const GalagoList *l;

	for (l = galago_service_get_accounts(service, TRUE);
		 l != NULL;
		 l = l->next)
	{
		append_account(menu, (GalagoAccount *)l->data);
	}
}

static void
rebuild_list(GalagoGtkAccountOptionMenu *optmenu)
{
	GtkWidget *menu;

	menu = gtk_menu_new();

	if (optmenu->priv->person != NULL)
	{
		append_accounts_from_person(optmenu, menu,
									optmenu->priv->person,
									optmenu->priv->service);
	}
	else if (optmenu->priv->person == NULL && optmenu->priv->me_person_set)
	{
		gtk_widget_destroy(menu);
		return;
	}
	else if (optmenu->priv->service != NULL)
	{
		append_accounts_from_service(optmenu, menu,
									 optmenu->priv->service);
	}
	else
	{
		const GalagoList *l;

		for (l = galago_core_get_services(FALSE, TRUE); l != NULL; l = l->next)
		{
			append_accounts_from_service(optmenu, menu,
										 (GalagoService *)l->data);
		}
	}

	gtk_widget_show(menu);

	gtk_option_menu_set_menu(GTK_OPTION_MENU(optmenu), menu);
	gtk_option_menu_set_history(GTK_OPTION_MENU(optmenu), 0);
}

static void
account_added_cb(void *obj, const char *property, GalagoAccount *account,
				 GalagoGtkAccountOptionMenu *optmenu)
{
	GalagoAccount *cur_account;

	cur_account = galago_gtk_account_option_menu_get_active(optmenu);

	rebuild_list(optmenu);

	if (cur_account != NULL)
	{
		if (galago_gtk_account_option_menu_get_active(optmenu) != cur_account)
			galago_gtk_account_option_menu_set_active(optmenu, cur_account);
	}
	else
		gtk_option_menu_set_history(GTK_OPTION_MENU(optmenu), 0);
}

static void
account_removed_cb(void *obj, const char *property, GalagoAccount *account,
				   GalagoGtkAccountOptionMenu *optmenu)
{
	GalagoAccount *cur_account;

	cur_account = galago_gtk_account_option_menu_get_active(optmenu);

	rebuild_list(optmenu);

	if (cur_account == account)
	{
		if (galago_gtk_account_option_menu_get_active(optmenu) != cur_account)
			galago_gtk_account_option_menu_set_active(optmenu, cur_account);
	}
	else
		gtk_option_menu_set_history(GTK_OPTION_MENU(optmenu), 0);
}

static void
person_added_cb(void *obj, const char *property, GalagoPerson *person,
				GalagoGtkAccountOptionMenu *optmenu)
{
	if (optmenu->priv->me_person_set && galago_person_is_me(person))
		galago_gtk_account_option_menu_set_person(optmenu, person);
}

static void
person_destroy_cb(GalagoPerson *person, const char *property,
				  GalagoGtkAccountOptionMenu *optmenu)
{
	optmenu->priv->person = NULL;

	galago_signals_disconnect_by_id(optmenu->priv->person_destroy_id);
	optmenu->priv->person_destroy_id = 0;

	if (galago_person_is_me(person))
	{
		optmenu->priv->person_added_id =
			galago_signal_connect(galago_get_core(), "person-added",
								  GALAGO_CALLBACK(person_added_cb),
								  optmenu);
	}
}

GtkWidget *
galago_gtk_account_option_menu_new(gboolean me_accounts)
{
	GalagoGtkAccountOptionMenu *optmenu;

	optmenu = g_object_new(GALAGO_GTK_TYPE_ACCOUNT_OPTION_MENU, NULL);

	optmenu->priv->me_accounts = me_accounts;

	if (me_accounts)
	{
		galago_gtk_account_option_menu_set_person(optmenu,
			galago_core_get_me(FALSE, TRUE));
	}

	rebuild_list(optmenu);

	gtk_option_menu_set_history(GTK_OPTION_MENU(optmenu), 0);

	return GTK_WIDGET(optmenu);
}

void
galago_gtk_account_option_menu_set_person(GalagoGtkAccountOptionMenu *optmenu,
										GalagoPerson *person)
{
	g_return_if_fail(optmenu != NULL);
	g_return_if_fail(GALAGO_GTK_IS_ACCOUNT_OPTION_MENU(optmenu));

	if (optmenu->priv->person == person)
		return;

	if (optmenu->priv->person_added_id > 0)
	{
		galago_signals_disconnect_by_id(optmenu->priv->person_added_id);
		optmenu->priv->person_added_id = 0;
	}

	if (optmenu->priv->person_destroy_id > 0)
	{
		galago_signals_disconnect_by_id(optmenu->priv->person_destroy_id);
		optmenu->priv->person_destroy_id = 0;
	}

	optmenu->priv->person = person;

	if (person != NULL)
	{
		optmenu->priv->me_person_set = galago_person_is_me(person);

		optmenu->priv->person_destroy_id =
			galago_signal_connect(person, "destroy",
								  GALAGO_CALLBACK(person_destroy_cb),
								  optmenu);
	}
	else
		optmenu->priv->me_person_set = FALSE;

	rebuild_list(optmenu);
}

GalagoPerson *
galago_gtk_account_option_menu_get_person(
	const GalagoGtkAccountOptionMenu *optmenu)
{
	g_return_val_if_fail(optmenu != NULL,                            NULL);
	g_return_val_if_fail(GALAGO_GTK_IS_ACCOUNT_OPTION_MENU(optmenu), NULL);

	return optmenu->priv->person;
}

void
galago_gtk_account_option_menu_set_service(GalagoGtkAccountOptionMenu *optmenu,
										 GalagoService *service)
{
	g_return_if_fail(optmenu != NULL);
	g_return_if_fail(GALAGO_GTK_IS_ACCOUNT_OPTION_MENU(optmenu));

	if (optmenu->priv->service == service)
		return;

	if (optmenu->priv->account_added_id > 0)
	{
		galago_signals_disconnect_by_id(optmenu->priv->account_added_id);
		optmenu->priv->account_added_id = 0;
	}

	if (optmenu->priv->account_removed_id > 0)
	{
		galago_signals_disconnect_by_id(optmenu->priv->account_removed_id);
		optmenu->priv->account_removed_id = 0;
	}

	optmenu->priv->service = service;

	if (service == NULL)
	{
		optmenu->priv->account_added_id =
			galago_signal_connect_class(GALAGO_CLASS_SERVICE, "account-added",
										GALAGO_CALLBACK(account_added_cb),
										optmenu);
		optmenu->priv->account_removed_id =
			galago_signal_connect_class(GALAGO_CLASS_SERVICE,
										"account-removed",
										GALAGO_CALLBACK(account_removed_cb),
										optmenu);
	}
	else
	{
		optmenu->priv->account_added_id =
			galago_signal_connect(service, "account-added",
								  GALAGO_CALLBACK(account_added_cb),
								  optmenu);
		optmenu->priv->account_removed_id =
			galago_signal_connect(service, "account-removed",
								  GALAGO_CALLBACK(account_removed_cb),
								  optmenu);
	}

	rebuild_list(optmenu);
}

GalagoService *
galago_gtk_account_option_menu_get_service(
	const GalagoGtkAccountOptionMenu *optmenu)
{
	g_return_val_if_fail(optmenu != NULL,                            NULL);
	g_return_val_if_fail(GALAGO_GTK_IS_ACCOUNT_OPTION_MENU(optmenu), NULL);

	return optmenu->priv->service;
}

void
galago_gtk_account_option_menu_set_active(GalagoGtkAccountOptionMenu *optmenu,
										const GalagoAccount *account)
{
	GtkWidget *menu;
	GList *l;
	int i;

	g_return_if_fail(optmenu != NULL);
	g_return_if_fail(GALAGO_GTK_IS_ACCOUNT_OPTION_MENU(optmenu));
	g_return_if_fail(account != NULL);

	menu = gtk_option_menu_get_menu(GTK_OPTION_MENU(optmenu));

	for (l = gtk_container_get_children(GTK_CONTAINER(menu)), i = 0;
		 l != NULL;
		 l = l->next, i++)
	{
		GtkWidget *menuitem = GTK_WIDGET(l);
		GalagoAccount *temp_account;

		if (!GALAGO_GTK_IS_ACCOUNT_MENU_ITEM(menuitem))
			continue;

		temp_account = galago_gtk_account_menu_item_get_account(
			GALAGO_GTK_ACCOUNT_MENU_ITEM(menuitem));

		if (temp_account == account)
		{
			gtk_option_menu_set_history(GTK_OPTION_MENU(optmenu), i);
			break;
		}
	}
}

GalagoAccount *
galago_gtk_account_option_menu_get_active(
	const GalagoGtkAccountOptionMenu *optmenu)
{
	GtkWidget *menu;
	GtkWidget *menuitem;

	g_return_val_if_fail(optmenu != NULL,                            NULL);
	g_return_val_if_fail(GALAGO_GTK_IS_ACCOUNT_OPTION_MENU(optmenu), NULL);

	menu = gtk_option_menu_get_menu(GTK_OPTION_MENU(optmenu));

	if (menu == NULL)
		return NULL;

	menuitem = gtk_menu_get_active(GTK_MENU(menu));

	g_return_val_if_fail(GALAGO_GTK_IS_ACCOUNT_MENU_ITEM(menuitem), NULL);

	return galago_gtk_account_menu_item_get_account(
		GALAGO_GTK_ACCOUNT_MENU_ITEM(menuitem));
}

static void
service_changed_cb(GalagoGtkServiceOptionMenu *service_optmenu,
				   GalagoService *service, GalagoGtkAccountOptionMenu *optmenu)
{
	galago_gtk_account_option_menu_set_service(optmenu,
		galago_gtk_service_option_menu_get_active(service_optmenu));
}

static void
service_option_menu_destroyed_cb(GalagoGtkServiceOptionMenu *service_optmenu,
							   GalagoGtkAccountOptionMenu *optmenu)
{
	optmenu->priv->service_changed_sig_id               = 0;
	optmenu->priv->service_option_menu_destroyed_sig_id = 0;
	optmenu->priv->service_optmenu                      = NULL;

	galago_gtk_account_option_menu_set_service(optmenu, NULL);
}

void
galago_gtk_account_option_menu_set_services_option_menu(
	GalagoGtkAccountOptionMenu *optmenu, GtkWidget *service_optmenu)
{
	GalagoService *service = NULL;

	g_return_if_fail(optmenu != NULL);
	g_return_if_fail(GALAGO_GTK_IS_ACCOUNT_OPTION_MENU(optmenu));
	g_return_if_fail(service_optmenu == NULL ||
					 GALAGO_GTK_IS_SERVICE_OPTION_MENU(service_optmenu));

	if (service_optmenu == optmenu->priv->service_optmenu)
		return;

	if (optmenu->priv->service_changed_sig_id != 0)
	{
		g_signal_handler_disconnect(
			G_OBJECT(optmenu->priv->service_optmenu),
			optmenu->priv->service_changed_sig_id);
	}

	if (optmenu->priv->service_option_menu_destroyed_sig_id != 0)
	{
		g_signal_handler_disconnect(
			G_OBJECT(optmenu->priv->service_optmenu),
			optmenu->priv->service_option_menu_destroyed_sig_id);
	}

	optmenu->priv->service_optmenu = service_optmenu;

	if (service_optmenu != NULL)
	{
		optmenu->priv->service_changed_sig_id =
			g_signal_connect(G_OBJECT(service_optmenu), "service_changed",
							 G_CALLBACK(service_changed_cb), optmenu);
		optmenu->priv->service_option_menu_destroyed_sig_id =
			g_signal_connect(G_OBJECT(service_optmenu), "destroy",
							 G_CALLBACK(service_option_menu_destroyed_cb),
							 optmenu);

		service = galago_gtk_service_option_menu_get_active(
			GALAGO_GTK_SERVICE_OPTION_MENU(service_optmenu));
	}

	galago_gtk_account_option_menu_set_service(optmenu, service);
}

GtkWidget *
galago_gtk_account_option_menu_get_services_option_menu(
	const GalagoGtkAccountOptionMenu *optmenu)
{
	g_return_val_if_fail(optmenu != NULL,                            NULL);
	g_return_val_if_fail(GALAGO_GTK_IS_ACCOUNT_OPTION_MENU(optmenu), NULL);

	return optmenu->priv->service_optmenu;
}
